"""
Init file for image module to facilitate importation
"""

__all__ = [
    "PointsRender",
    "CirclesRender",
    "EllipsesRender",
    "PolygonsRender",
    "SegmentsRender",
    "LinearSplinesRender",
    "OcrSingleOutputRender",
    "Image",
    "interpolate_color",
]

from otary.image.components.drawer.utils.render import (
    PointsRender,
    CirclesRender,
    EllipsesRender,
    PolygonsRender,
    SegmentsRender,
    LinearSplinesRender,
    OcrSingleOutputRender,
)
from otary.image.image import Image
from otary.image.utils.colors import interpolate_color
